-- rslib51
-- Version 2023-03-19

-- These routines were originally written for Euphoria Version 3.1, but I'm using them now with Phix
-- Slowly adapting them to Phix, though from the current Phix POV some things probably are outdated, some redundant, others may need to be updated ...
-- Phix: http://phix.x10.mx/

-- rslib 44: some functions renamed, because their names duplicated those of Euphoria/Phix routines (which hadn't existed at the time I wrote mine, long ago)
-- I keep mine because they work slightly differently, and may not be fully compatible
-- If you have used rslib versions prior to 44, call the renamed functions ('a' for alternative), or make sure the Euphoria/Phix ones work the way you want them
--   atrim instead of trim
--   asplit instead of split
--   ajoin instead of join

-- *** IMPORTANT ***
-- rslib 442: function begins() deleted because it is now part of Phix, but in the Phix function, parameters are in reverse order than they had been here!

-- I have written these routines for my own use, they are subject to change, their documentation is fragmentary, their functionality is not guaranteed.
-- This file is mainly included for the purpose of documentation. You are free to use this library, or create modified versions for your own purposes, but do so at your own peril.
-- Robert Schaechter, www.hypatia-rpn.net


-- function findeggs()
-- function findfiles(objecf f)
-- function findsfiles(object f)
-- function cut(object s, integer c)
-- function atrim(object s)
-- function rtrim(object s)
-- function asplit(object s, object sep)
-- function splitb(object s)
-- function splitn(object s, object sep)
-- function split2(object s, object sep)
-- function splitx(object s, object sep)
-- function ajoin(object s, atom sep)
-- function repl(object s, object sf, object sr)
-- function brack(object d, object s)
-- function readfile(object d)
-- function clipcopy(object f)
-- function yesno()
-- function sprint12(atom a)
-- procedure disp(object d)
-- procedure writefile(object s, object d)


-- note:
-- "substrings" and "set of substrings" refers to sequences in which each element is a "string"


-- include get.e
-- include misc.e
-- include file.e
-- include sort.e
-- include wildcard.e


-- finds files whose names start with @
-- returns their names, and their names with @ stripped (names are sorted alphabetically)
global function findeggs()
	object x, y, d
	x = ""
	d = dir(current_dir())
	for j = 1 to length(d) do
		if begins("@", d[j][1]) then
			y = ""
			y = append(y, d[j][1])
			y = append(y, d[j][1][2..length(d[j][1])])
			x = append(x, y)
		end if
	end for
	x = sort(x)
	return x
end function


-- splits sequence s in two parts at position c: 1 to c / c+1 to end
-- with c negative: c characters from end
global function cut(object s, integer c)
	object x
	x = ""
	if c < 0 then
		c = length(s) + c
		if c < 0 then c = 0 end if
	end if
	if c > length(s) then c = length(s) end if
	x = append(x, s[1..c])
	x = append(x, s[c+1..length(s)])
	return x
end function


-- strips leading and trailing characters with ASCII codes <= 32 from sequence s,
-- and also from its elements, if these are sequences
global function atrim(object s)
	integer first
	integer last
	first = 0
	last = 0
	for j = 1 to length(s) do
		if atom(s[j]) then
			if s[j] > 32 then
				last = j
				if first = 0 then first = j end if
			end if
		else
			last = j
			if first = 0 then first = j end if
			s[j] = atrim(s[j])
		end if
	end for
	if first then
		s = s[first..last]
	else
		s = {}
	end if
	return s
end function


-- strips trailing characters with ASCII codes <= 32 from sequence s,
-- and also from its elements, if these are sequences
global function rtrim(object s)
	integer last
	last = 0
	for j = 1 to length(s) do
		if atom(s[j]) then
			if s[j] > 32 then last = j end if
		else
			last = j
			s[j] = rtrim(s[j])
		end if
	end for
	if last then
		s = s[1..last]
	else
		s = {}
	end if
	return s
end function


-- splits string s into substrings at the occurences of sep
-- if sep is an atom, it will not be included in the substrings
-- if sep is a sequence, it will be included
global function asplit(object s, object sep)
	object feld
	object x
	integer a
	integer i
	feld = ""
	x = ""
	if atom(sep) then
		for j = 1 to length(s) do
			if s[j] = sep then
				x = append(x, feld)
				feld = ""
			else
				feld = feld & s[j]
			end if
		end for
	else
		a = length(sep) - 1
		i = 1
		while i < length(s) - a do
			if equal(s[i..i+a], sep) then
				x = append(x, feld & sep)
				feld = ""
				i = i + a + 1
			else
				feld = feld & s[i]
				i += 1
			end if
		end while
	end if
	x = append(x, feld)
	return x
end function


-- splits string s into substrings, separated at space characters
-- several consecutive space characters are treated as one
global function splitb(object s)
	object feld
	object x
	integer a
	feld = ""
	x = ""
	a = 1

	for j = 1 to length(s) do
		if s[j] = ' ' then
			if a then
				x = append(x, feld)
				feld = ""
			end if
			a = 0
		else
			feld = feld & s[j]
			a = 1
		end if
	end for

	x = append(x, feld)
	return x
end function


-- splits string s into n substrings, separated at the occurence of sep, which must be a sequence
global function splitn(object s, object sep)
	object x
	integer a
	x = {}

	while length(s) do
		a = match(sep, s)
		if a then
			x = append(x, s[1..a-1])
			s = s[a+2..$]
		else
			x = append(x, s)
			exit
		end if
	end while
	return x
end function


-- splits string s into two substrings at the first occurence of sep (which must be an atom)
-- if sep is not found, the first returned substring is empty
global function split2(object s, atom sep)
	object x
	integer n
	x = {"",""}
	n = 0
	for j=1 to length(s) do
		if s[j] = sep then
			n = j
			exit
		end if
	end for

	if n = 0 then
		x[2] = s
	else

		x[1] = s[1..n-1]
		x[2] = s[n+1..$]
	end if
	return x
end function


-- splits string s into two substrings at the last occurence of sep (which must be an atom)
-- if sep is not found, the second returned substring is empty
global function splitx(object s, atom sep)
	object x
	integer n
	x = {"",""}
	n = 0
	for j= length(s) to 1 by -1 do
		if s[j] = sep then
			n = j
			exit
		end if
	end for

	if n = 0 then
		x[1] = s
	else

		x[1] = s[1..n-1]
		x[2] = s[n+1..$]
	end if
	return x
end function


-- joins substrings to one string
-- if sep = 0, no characters are inserted between substrings
-- if sep = ' ' then space characters are inserted if substring lines are not empty
-- with other values of sep, this atom is inserted between all substrings, empty or not
global function ajoin(object s, atom sep)
	object x
	x = ""
	for j = 1 to length(s) do
		if sep and j > 1 then
			if sep = ' ' then
				if length(s[j-1]) then x = x & ' ' end if
			else
				x = x & sep
			end if
		end if
		x = x & s[j]
	end for
	return x
end function


-- replaces substring sf with substring sr within string s
-- this is not the same as the Euphoria/Phix replace function
global function repl(object s, object sf, object sr)
	integer i, l
	i = 1
	l = length(sf)

	while i do
		i = match(sf, s, i)
		if i then
			s = s[1..i-1] & sr & s[i+l..$]
		end if
	end while
	return s
end function


-- reduces text within brackets to single elements of a string
-- the result is not a string anymore, as its elements contain sequences!
-- s[1] is opening bracket, s[2] is closing bracket (must be atoms)
global function brack(object d, object s)
	sequence x, x1
	integer flag
	x = ""
	x1 = ""
	flag = 0
	for j = 1 to length(d) do
		if flag then
			x1 = append(x1, d[j])
			if d[j] = s[2] then
				flag = 0
				x = append(x, x1)
				x1 = ""
			end if
		else
			if d[j] = s[1] then
				flag = 1
				x1 = append(x1, d[j])
			else
				x = append(x, d[j])
			end if
		end if
	end for
	return x
end function


-- used in findsfiles()
function tree()
	object subs
	object d
	integer ia, io

	subs = ""
	d = dir(current_dir())
	if atom(d) then return subs end if

	if equal(d[1][1], ".") then ia = 3 else ia = 1 end if
	for j = ia to length(d) do
		if equal(d[j][2], "d") then subs = append(subs, d[j][1]) end if
	end for

	ia = 1
	io = length(subs)
	if io = 0 then return subs end if

	while 1 do
		for i = ia to io do
			d = dir(subs[i])
			for j = 3 to length(d) do
				if equal(d[j][2], "d") then subs = append(subs, subs[i] & "\\" & d[j][1]) end if
			end for
		end for
		if length(subs) = io then exit end if

		ia = io + 1
		io = length(subs)
	end while

	subs = sort(subs)
	return subs
end function


-- finds files whose names match the search string (may contain wildcard characters)
-- search string may be sequence of strings
-- returns their names, sorted alphabetically
global function findfiles(object f)
	object x, d, y
	x = {}
	if atom(f[1]) then f = {f} end if

	for i = 1 to length(f) do
		d = dir(f[i])
		if not atom(d) then
			for j = 1 to length(d) do
				if not equal (d[j][2], "d") then
					x = append(x, d[j][1])
				end if
			end for
		end if
	end for

	for i = 1 to length(x) do
		y = x[i]
		x[i] = upper(x[i])
		x[i] = splitx(x[i], '.')
		x[i] = append(x[i], y)
	end for
	
	x = sort(x)

	for i = 1 to length(x) do
		x[i] = x[i][3]
	end for

	return x
end function


-- as findfiles, but includes subdirectories
-- CURRENT VERSION NOT TESTED!!!
global function findsfiles(object f)
	object subs  -- list of subdirectories
	object x, xx, y
	if atom(f[1]) then f = {f} end if

	subs = tree()
	x = findfiles(f)

	for j = 1 to length(subs) do
		-- for Unix: slash, not backslash!
		for i = 1 to length(f) do
			xx = findfiles(subs[j] & "\\" & f[i])
			for jj = 1 to length(xx) do
				x = append(x, subs[j] & "\\" & xx[jj])
			end for
		end for
	end for

	for i = 1 to length(x) do
		y = x[i]
		x[i] = upper(x[i])
		x[i] = splitx(x[i], '.')
		x[i] = append(x[i], y)
	end for
	
	x = sort(x)

	for i = 1 to length(x) do
		x[i] = x[i][3]
	end for

	return x
end function


-- reads text file d line by line into a set of substrings
-- returns -1 if file doesn't exist
global function readfile(object d)
	integer Din
	object line
	object s
	s = {}
	Din = open(d, "r")
	if Din = -1 then
		s = -1
		return s
	else
		while 1 do
			line = gets(Din)
			if atom(line) then
				close(Din)
				s = rtrim(s)
				return s
			end if
			s = append(s, line)
		end while
	end if
end function


-- copies content of file to the Windows clipboard
-- returns 1 if file found, 0 if not found
global function clipcopy(object f)
	object d
	d = dir(f)
	if atom(d) then
		return 0
	else
		system("clip < " & f, 2)
		return 1
	end if
end function


-- prompts user to press Y or N key (or ENTER or ESC key)
-- returns 1 for Y or ENTER, 0 for N or ESC
global function yesno()
	integer yes
	integer ky
	yes = -1
	puts(1, "(Y/N or Enter/Esc) ")
	while yes = -1 do
		ky = wait_key()
		if ky = 27 or ky = 78 or ky = 110 then
			yes = 0
		elsif ky = 13 or ky = 89 or ky = 121 then
			yes = 1
		end if
	end while
	puts(1, "\n")
	return yes
end function


-- a substitute for sprint(), returns up to 12 digits instead of 10
-- like sprint(), switches to scientific (exponential) notation for small and large numbers,
-- but unlike sprint() still returns up to 12 significant digits after (up to 3) zeros after decimal point for absolute values < 1
global function sprint12(atom a)

	string f, g
	integer flag_neg = 0
	integer pos_e
	atom xp

	if a = 0 then
		return "0"
	elsif a > 0 then
		flag_neg = 0
	else
		a = -a
		flag_neg = 1
	end if

	f = sprintf("%1.11e", a)            -- creates string "1.23456789012exx"
	pos_e = match('e', f)               -- position of 'e' in string f (should be 14, but can be 15)
	
	if pos_e < 14 or pos_e > 15 then    -- to be on the safe side
		return sprint(a)
	end if
	
	g = f[1] & f[3..13]                 -- creates string "123456789012"
	xp = value(f[pos_e+1..$])[2]        -- value of exponent

	if xp < -4 or xp >= 12 then
		g = f[1..13]
		while g[$] = '0' do
			g = g[1..$-1]
		end while
		if g[$] = '.' then g = g[1..$-1] end if
		g = g & f[pos_e..$]
	elsif xp < 0 then
		g = "0." & repeat('0', -xp-1) & g
		while g[$] = '0' do
			g = g[1..$-1]
		end while
	else
		g = g[1..xp+1] & '.' & g[xp+2..$]
		while g[$] = '0' do
			g = g[1..$-1]
		end while
		if g[$] = '.' then g = g[1..$-1] end if
	end if

	if flag_neg then g = '-' & g end if
	return g
end function


-- simple shortcut for puts command to write a line of text to the screen (stdout)
global procedure disp(object d)
	puts(1, d & "\n")
end procedure


-- writes set of substrings to file d
global procedure writefile(object s, object d)
	integer Dout
	Dout = open(d, "w")
	if Dout = -1 then
		puts(1, "Error: could not write file " & d & "\n(Invalid file name, or write protected)\n")
		abort(1)
	else
		for j = 1 to length(s) do
			puts(Dout, s[j] & "\n")
		end for
		close(Dout)
	end if
end procedure